/* 
   $Id: calendar.c,v 1.5 2001/12/27 08:21:15 sho Exp $

   Copyright (C) 2001,2002 AXE,Inc. <sg@axe-inc.co.jp>

   This file is part of sgsched.

   sgsched is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   sgsched is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with sgsched; see the file COPYING.   If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <unistd.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#include "calendar.h"

static  struct dayrule  *holidays;
static  struct dayrule  *memorials;

static int    get_nth_dayweek(int day);
static int    diff_date(struct date *a, struct date *b);

static struct dayrule * load_dayrule(char *fn);
static struct dayrule * comp_dayrule_by_date(struct dayrule *rule, int year, int month, int day);
static struct date    * comp_dayrule_by_name(struct dayrule *rule, char *name, int year);
static struct date    * make_holiday(struct dayrule *h, int year, int month, int day);

static struct dayrule *load_dayrule(char *fn)
{
  FILE       *fp;
  char       t[BUFSIZ];
  char       *p;
  int        c, j;

  struct dayrule    *top = NULL, *cur = NULL, tmp;

  if(fn == NULL) return 0;

  fp = fopen(fn, "r"); 
  if(fp == NULL){
    fprintf(stderr, "I can't open file:%s.\n",fn);
    return NULL;
  }

  p = t, j = 0;

  top = (struct dayrule *) calloc(1, sizeof(struct dayrule));
  cur = top;

  while( (c = fgetc(fp)) != EOF){
    switch(c){
    case '#':
      while( (c = fgetc(fp)) != EOF && c != '\n');
      break;

    case '"':
      while( (c = fgetc(fp)) != EOF && c != '"') *p++ = c;
      break;

    case ' ':
    case '\n':
    case '\t':
    case '\0':
      if(p != t){
	*p = '\0';

	switch(j){
	case 0:
	  tmp.name = (char *) strdup(t);
	  break;
	case 1:
	  if(t[0] == '*') tmp.month = -1;
	  else            tmp.month = atoi(t);
	  break;
	case 2:
	  if(t[0] == '*') tmp.mday = -1;
	  else            tmp.mday = atoi(t);
	  break;
	case 3:
	  if(t[0] == '*') tmp.nth = -1;
	  else            tmp.nth = atoi(t);
	  break;
	case 4:
	  if(t[0] == '*') tmp.wday = -1;
	  else            tmp.wday = atoi(t);
	  break;
	case 5:
	  if(t[0] == '*') tmp.equinox = -1;
	  else            tmp.equinox = atoi(t);
	  break;
	}

	p = t;
	j++;
	if(j > 5){
	  cur->next = (struct dayrule *) calloc(1, sizeof(struct dayrule));
	  cur       = cur->next;
	  *cur      = tmp;
	  cur->next = NULL;
	  j = 0;
	}
      }
      break;

    default:
      *p++ = c;
      break;
    }
  }

  if(p != t){
    *p = '\0';

    if(j == 5){
      if(t[0] == '*') tmp.equinox = -1;
      else            tmp.equinox = atoi(t);

      cur->next = (struct dayrule *) calloc(1, sizeof(struct dayrule));
      cur       = cur->next;
      *cur      = tmp;
      cur->next = NULL;
    }
  }
  
  fclose(fp);
  return top;
}

int get_spring_equinox(int year)
{
  double d;
  int    i;

  i = (year - 2000) / 4;
  d = (20.69115 + ((year - 2000) * 0.242194) - i);
  
  return (int)d;
}

int get_autumn_equinox(int year)
{
  double d;
  int    i;

  i = (year - 2000) / 4;
  d = (23.09 + ((year - 2000) * 0.242194) - i);
  
  return (int)d;
}

int get_leapyear(int year)
{
  int i;
  i = (((year % 4) == 0) && ((year % 100) != 0)) || ((year % 400) == 0) ? 1 : 0;
  return i;
}

int get_day_of_month(int year, int month)
{
  int days[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
  if(month == 2 && get_leapyear(year)) return 29;
  return days[month - 1];
}

int get_dayweek(int year, int month, int day)
{
  int weekday;

  if(month <= 2) year--, month+=12;

  weekday = (year + (year / 4) - (year / 100)
	     + (year / 400) + (13 * month + 8) / 5 + day) % 7;

  return weekday;
}

static int get_nth_dayweek(int day)
{
  return (day / 7) + ((day % 7) ? 1 : 0);
}

int load_holiday(char *fn)
{
  holidays = load_dayrule(fn);
  if(holidays) return 1;
  else         return 0;
}

int load_memorial(char *fn)
{
  memorials = load_dayrule(fn);
  if(memorials) return 1;
  else         return 0;
}

static struct dayrule *
comp_dayrule_by_date(struct dayrule *rule, int year, int month, int day)
{
  int f = 0;
  int i;
  int nth, wday;
  struct dayrule *tmp;

  if (rule == NULL) return NULL;
  wday = get_dayweek(year, month, day);
  nth  = get_nth_dayweek(day);

  tmp = rule->next;

  i = 0;
  while(tmp){
    i++;
    f = 1;
    
    if(tmp->month != -1 && tmp->month != month) f = 0;
    if(tmp->mday  != -1 && tmp->mday  != day)   f = 0;
    if(tmp->nth   != -1 && tmp->nth   != nth)   f = 0;
    if(tmp->wday  != -1 && tmp->wday  != wday)  f = 0;

    if(tmp->equinox != -1){
      if(month == 3 && tmp->month == 3 && day == get_spring_equinox(year))       f = 1;
      else if(month == 9 && tmp->month == 9 && day == get_autumn_equinox(year))  f = 1;
      else                                                    f = 0;
    }

    if(f) return tmp;

    tmp = tmp->next;
  }

  return NULL;
}

static struct date *
comp_dayrule_by_name(struct dayrule *rule, char *name, int year)
{
  struct date *date = NULL;
  int y = year, m, d, w, n, e;

  if(!rule || rule->name == NULL)   return NULL;

  if(strcmp(rule->name, name) == 0){
    m = rule->month;
    d = rule->mday;
    w = rule->wday;
    n = rule->nth;
    e = rule->equinox;

    if(m != -1){
      if(d == -1){
	if(w != -1 && n != -1){
	  d = (n - 1) * 7;
	  while(get_dayweek(y, m, d) != w) d++;
	}else if((m == 3 || m == 9) && e != -1){
	  if(m == 3)      d = get_spring_equinox(y);
	  else if(m == 9) d = get_autumn_equinox(y);
	}
      }else{
	if(n != -1 && n != get_nth_dayweek(d))   return NULL;
	if(w != -1 && w != get_dayweek(y, m, d)) return NULL;
	w = get_dayweek(y, m, d);
      }

      if(w == 0){
	next_day(y, m, d);
      }

      date = (struct date *) calloc(1, sizeof(struct date));
      date->year = y;
      date->mon  = m;
      date->mday = d;
    }
  }

  return date;
}

int is_holiday(int year, int month, int day)
{
  int         wday;
  int         py, pm, pd;

  wday = get_dayweek(year, month, day);
  
  py = year, pm = month, pd = day;
  prev_day(py, pm, pd);

  if(wday == 0)
    return 1;
  else if(wday == 1 && comp_dayrule_by_date(holidays, py, pm, pd))
    return 1;
  else if(comp_dayrule_by_date(holidays, year, month, day))
    return 1;
  
  return 0;
}

static struct date *
make_holiday(struct dayrule *h, int year, int month, int day)
{
  struct date *date = NULL;
  int y = year, m, d, w, n, e;

  if(!h) return NULL;

  m = h->month;
  d = h->mday;
  w = h->wday;
  n = h->nth;
  e = h->equinox;

  if(m != -1){
    if(d == -1){
      if(w != -1 && n != -1){
	d = (n - 1)* 7 + w;
	while(get_dayweek(y, m, d) != w) d++;
      }else if((m == 3 || m == 9) && e != -1){
	if(m == 3)      d = get_spring_equinox(y);
	else if(m == 9) d = get_autumn_equinox(y);
	w = get_dayweek(y, m, d);
      }
    }else{
      if(n != -1 && n != get_nth_dayweek(d))   return NULL;
      if(w != -1 && w != get_dayweek(y, m, d)) return NULL;
      w = get_dayweek(y, m, d);
    }
    
    date = (struct date *) calloc(1, sizeof(struct date));
    
    date->year = y;
    date->mon  = m;
    date->mday = d;
  }

  return date;
}

struct date *
get_next_sunday(int year, int month, int day)
{
  struct  date *date;
  int w;

  w = get_dayweek(year, month, day);

  if((day + (7 - w)) > get_day_of_month(year, month)){
    day = (day + (7 - w)) - get_day_of_month(year, month);
    month++;
    if(month > 12) year++, month = 1;
  }else{
    day += (7 - w);
  }

  date = (struct date *) calloc(1, sizeof(struct date));

  date->year = year;
  date->mon  = month;
  date->mday = day;

  return date;
}

struct date *
get_next_public_holiday(int year, int month, int day)
{
  struct  dayrule      *tmp;
  int          nth, wday;
  struct  date *h, *r = NULL;
  struct  date t;

  if(!holidays) return NULL;

  wday = get_dayweek(year, month, day);
  nth  = get_nth_dayweek(day);

  if(wday == 0){
    if(comp_dayrule_by_date(holidays, year, month, day)){
      r = (struct date *) calloc(1, sizeof(struct date));
    
      r->year = year;
      r->mon  = month;
      r->mday = day;
      next_day(r->year, r->mon, r->mday);
      return r;
    }
  }
  
  tmp = holidays->next;

  t.year  = year;
  t.mon   = month;
  t.mday  = day;

  while(tmp){
    h = make_holiday(tmp, year, month, day);

    if(h){
      if(comp_date(&t, h) < 0){
	r = h;
	break;
      }
    }
    
    free(h);
    tmp = tmp->next;
  }

  if(r == NULL) 
    r = get_next_public_holiday(year + 1, 1, 0);
  
  return r;
}

int comp_date(struct date *a, struct date *b)
{
  if(a->year > b->year)      return  1;
  else if(a->year < b->year) return -1;
  else{
    if(a->mon > b->mon)      return  1;
    else if(a->mon < b->mon) return -1;
    else{
      if(a->mday > b->mday)      return  1;
      else if(a->mday < b->mday) return -1;
      else                       return  0;
    }
  }
}

time_t get_time_t(int year, int mon, int day, int hour, int min, int sec)
{
  struct tm tm;

  tm.tm_year  = year - 1900;
  tm.tm_mon   = mon - 1;
  tm.tm_mday  = day;
  tm.tm_hour  = hour;
  tm.tm_min   = min;
  tm.tm_sec   = sec;
  tm.tm_isdst = -1;

  return mktime(&tm);
}

int diff_time_t(time_t t1, time_t t2)
{
  return  ((t2 - t1)/86400);
}

static int diff_date(struct date *a, struct date *b)
{
  time_t t1, t2;
  int d;
  
  t1 = get_time_t(a->year, a->mon, a->mday, 0, 0, 0);
  t2 = get_time_t(b->year, b->mon, b->mday, 0, 0, 0);

  d = diff_time_t(t1, t2);

  return d;
}

struct date *
get_next_holiday(int year, int month, int day)
{
  int             nth, wday;
  struct dayrule  *tmp;
  struct  date    *h, *s, *r = NULL;
  struct  date    d;
  struct  tm      *tm;
  time_t          t;

  if(!holidays) return NULL;

  wday = get_dayweek(year, month, day);
  nth  = get_nth_dayweek(day);

  tmp = holidays->next;

  time(&t);
  tm = localtime(&t);

  if(year == -1)  d.year = tm->tm_year + 1900;
  else            d.year = year;
  if(month == -1) d.mon  = tm->tm_mon + 1;
  else            d.mon  = month;
  if(day == -1)   d.mday = tm->tm_mday;
  else            d.mday = day;

  s = get_next_sunday(d.year, d.mon, d.mday);
  
  while(tmp){
    h = make_holiday(tmp, d.year, d.mon, d.mday);

    if(h){
      if(comp_date(&d, h) < 0){
	if(comp_date(s, h) > 0){
	  free(s);
	  r = h;
	  break;
	}else{
	  free(h);
	  r = s;
	  break;
	}
      }
    }

    free(h);
    tmp = tmp->next;
  }

  if(r == NULL) r = get_next_holiday(d.year + 1, 1, 0);
  
  return r;
}

struct date *
get_holiday_by_name(char *name, int year)
{
  struct dayrule *tmp;
  struct date     *d;
  struct tm       *tm;
  time_t          t;

  time(&t);
  tm = localtime(&t);

  if(year == -1)  year = tm->tm_year + 1900;

  tmp = holidays->next;

  while(tmp){
    d = comp_dayrule_by_name(tmp, name, year);
    if(d) return d;
    tmp = tmp->next;
  }

  return NULL;
}

char *
get_holiday_by_date(int year, int month, int day)
{
  struct dayrule *tmp;

  tmp = comp_dayrule_by_date(holidays, year, month, day);

  if(tmp == NULL){
    /*    prev_day(year, month, day);
    if(comp_dayrule_by_date(holidays, year, month, day))
      return "Substitute Holiday";
    else*/ return NULL;
  }
  else            return tmp->name;
}

/* nϢ٤õ */
struct date *
get_nholidays(int year, int month, int day, int n, int max)
{
  struct date *a, *b, *ret;
  struct date d, s;
  int    c;

  c = 0;
  s.year = year;
  s.mon  = month;
  s.mday = day;

  a = get_next_holiday(year, month, day), d = *a;
  while(c < n - 1){
    b = get_next_holiday(a->year, a->mon, a->mday);

    if(diff_date(a, b) == 1) c++;
    else                    c = 0, d = *b;

    free(a);
    a = b;

    if(diff_date(&s, a) > max){
      free(a);
      return NULL;
    }
  }    

  ret = (struct date *) calloc(1, sizeof(struct date));
  *ret = d;

  return ret;
}

struct date *
get_memorialday_by_name(char *name, int year)
{
  struct dayrule *tmp;
  struct date     *d;
  struct tm       *tm;
  time_t          t;

  if(!memorials) return NULL;

  time(&t);
  tm = localtime(&t);

  if(year == -1)  year = tm->tm_year + 1900;

  tmp = memorials->next;

  while(tmp){
    d = comp_dayrule_by_name(tmp, name, year);
    if(d) return d;
    tmp = tmp->next;
  }

  return NULL;
}

char *
get_memorialday_by_date(int year, int month, int day)
{
  struct dayrule *tmp;

  tmp = comp_dayrule_by_date(memorials, year, month, day);

  if(tmp == NULL) return NULL;
  else            return tmp->name;
}

struct date *
get_next_memorialday(int year, int month, int day)
{
  struct  dayrule      *tmp;
  int          nth, wday;
  struct  date *h, *r = NULL;
  struct  date t;

  if(!memorials) return NULL;

  wday = get_dayweek(year, month, day);
  nth  = get_nth_dayweek(day);

  tmp = memorials->next;

  t.year  = year;
  t.mon   = month;
  t.mday  = day;

  while(tmp){
    h = make_holiday(tmp, year, month, day);

    if(h){
      if(comp_date(&t, h) < 0){
	r = h;
	break;
      }
    }
    
    free(h);
    tmp = tmp->next;
  }

  if(r == NULL) 
    r = get_next_memorialday(year + 1, 1, 0);
  
  return r;
}
